(* Wolfram Language package *)

BeginPackage["QuantumAlgebra`RulesSupplement`", 
	{ 
		"QuantumAlgebra`", 
		"QuantumAlgebra`Constants`", 
		"QuantumAlgebra`Analyze`", 
		"QuantumAlgebra`Options`" 
	}]


(* ::Section:: *)
(* Exported Symbols *)

IntegrationConstantQ::usage = 
	"Check whether an expression is an integration constant"

VariableQ::usage = 
	"Check whether an expression is a valid variable"

QAIntegrate::usage = 
	"Performs Integration in the context of Quantum Algebra"

QASum::usage = 
	"Performs Summation in the context of Quantum Algebra"

QAD::usage = 
	"Performs Derivation in the context of Quantum Algebra"

QASelectDefault::usage = 
	"Select a default set of Assumptions"

QAResetExplicitIntegrationVariables::usage = 
	"Resets a given set of Explicit Integration Variables"

QAAddExplicitIntegrationVariables::usage = 
	"Adds to a given set of Explicit Integration Variables"

DefineCommutatorRule::usage =
	"Uses a Commutator / Anti-Commutator to sort operators"
	
DefineNameOrderingRule::usage =
	"Uses a Commutator / Anti-Commutator to sort operators according to name order"

QAReplace::usage = 
	"Replaces in an expression by applying a set of Replacement Rules"

QAReplaceRepeated::usage = 
	"Replaces repeatedly in an expression by applying a set of Replacement Rules"

QAInitializeReplacementRules::usage = 
	"Initializes a built-in set of Replacement Rules"

QAResetReplacementRules::usage =
	"Resets a given set of Replacement Rules"

QAAddReplacementRules::usage =
	"Adds to a given set of Replacement Rules"

QAResetAssumptions::usage = 
	"Resets a given set of Assumptions"

QAAddAssumptions::usage = 
	"Adds to a given set of Assumptions"

QASetTimeConstraint::usage =
	"Defines the Time Constraint for QASimplify (default: 300)"

QASimplify::usage = 
	"Performs Simplify using a given set of Assumptions"

RowQ::usage =
	"Checks whether a Row Matrix is given"

ColumnQ::usage =
	"Checks whether a Column Matrix is given"

MakeRow::usage =
	"Makes from a Vector a Row Matrix"

MakeColumn::usage =
	"Makes from a Vector a Column Matrix"


(* ::Section:: *)
(* Implementation of the Package *)

Begin["`Private`"]


(* ::Subsection:: *)
(* Predicates *)

(* Constant Definition *)
IntegrationConstantQ[ex_, x__, min_Integer, max_Integer] :=
	Module[{symbolsEqualQ, freeExplicitQ},
		symbolsEqualQ[s1_Symbol, s2_Symbol] := Order[SymbolName[s1], SymbolName[s2]] == 0;
		freeExplicitQ[ex2_, explicit: Subscript[name_Symbol, idx_Integer], ivar_Symbol] :=
			FreeQ[ex2, explicit] || !symbolsEqualQ[name, ivar];
		freeExplicitQ[ex2_, explicit: Superscript[name_Symbol, idx_Integer], ivar_Symbol] :=
			FreeQ[ex2, explicit] || !symbolsEqualQ[name, ivar];
		freeExplicitQ[ex2_, explicit_, ivar_Symbol] :=
			True;
		NumericQ[ex] ||
		And@@Flatten[Outer[(FreeQ[ex, Subscript[#1, #2]] && FreeQ[ex, Superscript[#1, #2]])&, {x}, Range[min, max]]] &&
		And@@Flatten[Outer[freeExplicitQ[ex, #2, #1]&, { x }, variables[getSet[Null]]]]
	]


(* Variable Name *)
VariableQ[x_] := MatchQ[x, _Symbol] || MatchQ[x, OverBar[_Symbol]] || MatchQ[x, OverTilde[_Symbol]]


(* ::Subsection:: *)
(* Integration Rules *)
QAIntegrate[0, v: (_Symbol | _DirectedInfinity), _Integer, __Symbol] := 0


QAIntegrate[0., v: (_Symbol | _DirectedInfinity), _Integer, __Symbol] := 0


QAIntegrate[c_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] && ! c === 1 :=
	c QAIntegrate[1, v, n, x]
	
	 
QAIntegrate[1, v_Symbol, n_Integer, x__Symbol] /;
	! v === Infinity := 
	v^Length[{x}]


QAIntegrate[c_ p_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] :=
	c QAIntegrate[p, v, n, x]
	
	
QAIntegrate /: QAIntegrate[p_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol] QAIntegrate[q_, v: (_Symbol | _DirectedInfinity), n_Integer, y__Symbol] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QAIntegrate[p q, v, n, Sequence[x, y]]


(* Integrate about 2 integration variables where DiracDelta is used containing both of them *)
QAIntegrate[exp: Times[facs__, DiracDelta[__]], v: (_Symbol | _DirectedInfinity), n_Integer, Sequence[a_Symbol, b_Symbol]] /;
	Module[{i},
		MatchQ[exp, Times[__, DiracDelta[Sequence@@Table[(Subscript[a, i] - Subscript[b, i]), {i, 1, n}]]]] ||
		MatchQ[exp, Times[__, DiracDelta[Sequence@@Table[(Subscript[b, i] - Subscript[a, i]), {i, 1, n}]]]]
	] :=
	Module[{facsrepl, i},
		facsrepl = Sequence@@({facs} /. Table[Subscript[b, i] -> Subscript[a, i], {i, 0, n}]);
		QAIntegrate[Times[facsrepl], v, n, a]
	]


(* 2 helper functions for the next 3 rules *)
testFormM[exp_, n_Integer, x_Symbol] :=
	Module[{i},
		MatchQ[Apply[List, exp // FullSimplify], Table[If[i == 0, Subscript[x, i] * (Subscript[j_, i] - Subscript[k_, i]), Subscript[x, i] * (Subscript[k_, i] - Subscript[j_, i])], {i, 0, n}]]
	]

changeFormM[exp_, n_Integer, x_Symbol] /;
	testFormM[exp, n, x] := 
	Module[{list, entry, res, i, j, rev},
		rev[a_ - b_] := b - a;
		rev[a_] := a;
		list = Apply[List, exp // FullSimplify, {0, 1}];
		For[i = 2, i <= n + 1, i = i + 1,
			entry = list[[i]];
			For[j = 1, j <= 2, j = j + 1,
				entry[[j]] = rev[entry[[j]]]
			];			
			list[[i]] = Append[entry, -1]
		];
		res = Apply[Times, list, {1}];
		res = Apply[Plus, res];
		res
	]


(* Integrate about an Exponential resolved to DiracDelta *)
QAIntegrate[Power[E, (+I | -I) * exp_], Infinity, n_Integer, k_Symbol] /;
	Module[{i},
		MatchQ[exp // FullSimplify, Plus@@Table[Subscript[k, i] * (Subscript[x_, i] - Subscript[y_, i]), {i, 1, n}]]
	] :=
	Module[{held, pos},
		held = Hold[exp // FullSimplify // Evaluate];
		pos = 3 - Position[held, k][[1, 3]];
		DiracDelta@@(Apply[List, held, {0, 2}][[1, 1 ;; n, pos]]) * (2 * Pi) ^ n
	]

(* Integrate about an Exponential resolved to DiscreteDelta *)
QAIntegrate[Power[E, (I | -I) * exp_], v_Symbol, n_Integer, x_Symbol] /;
	! v === Infinity && 
	testFormM[exp, n, x] :=
	Module[{held, pos},
		held = Hold[changeFormM[exp, n, x] // Evaluate];
		pos = 4 - Position[held, x][[1, 3]];
		DiscreteDelta@@(Apply[List, held, {0, 2}][[1, 2 ;; n + 1, pos]]) * v
	]

QAIntegrate[Power[E, (fac: (I | -I)) * exp_], v_Symbol, n_Integer, x_Symbol] /;
	Module[{i},
		! v === Infinity && 
		MatchQ[exp // FullSimplify, Plus@@Table[If[i == 0, +1, -1] Subscript[x, i] * (Subscript[j_, i] + Subscript[k_, i]), {i, 0, n}]]
	] :=
	Module[{held, pos},
		held = Hold[exp // FullSimplify // Evaluate];
		pos = 4 - Position[held, x][[1, 3]];
		DiscreteDelta@@(Apply[List, held, {0, 2}][[1, 2 ;; n + 1, pos]]) * v * (Exp[fac (Apply[List, held, {0, 2}][[1, 1, pos - 1]]) Subscript[x, 0]])
	]
	

(* ::Subsection:: *)
(* Summation Rules *)
QASum[0, __Symbol] := 0


QASum[0., __Symbol] := 0


QASum[c_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] && ! c === 1 :=
	c QASum[1, x] 


QASum[c_ p_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] :=
	c QASum[p, x]


QASum /: QASum[p_, x__Symbol] QASum[q_, y__Symbol] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QASum[p q, Sequence[x, y]]


(* Summation about 2 summation variables where DiscreteDelta is used containing both of them *) 
QASum[exp: Times[facs__, dd: DiscreteDelta[fa_, la___]], Sequence[a_Symbol, b_Symbol]] /;
	Module[{i},
		MatchQ[{fa, la}, Table[(Subscript[a, i] - Subscript[b, i]), {i, 1, Length[dd]}]] ||
		MatchQ[{fa, la}, Table[(Subscript[b, i] - Subscript[a, i]), {i, 1, Length[dd]}]] ||
		MatchQ[{fa, la}, Table[(Subscript[a, i] + Subscript[b, i]), {i, 1, Length[dd]}]] ||
		MatchQ[{fa, la}, Table[(Subscript[b, i] + Subscript[a, i]), {i, 1, Length[dd]}]]
	] :=
	Module[{facsrepl, i, sgn, p, q},
		sgn[i1_] := If[! MatchQ[fa, p_ - q_] && i1 > 0, -1, +1];
		facsrepl = Sequence@@({facs} /. Table[Subscript[b, i] -> sgn[i] Subscript[a, i], {i, 0, Length[dd]}]);
		QASum[Times[facsrepl], a]
	]


QASum[Subscript[k_, i_], k_] /; 1 <= i <= 3 := 0


(* ::Subsection:: *)
(* Exchange order of Integration / Summation *)
QAIntegrate[QASum[ex_, svar__Symbol], v: (_Symbol | _DirectedInfinity), n_Integer, ivar__Symbol] := QASum[QAIntegrate[ex, v, n, ivar], svar]


(* ::Subsection:: *)
(* Integration Rules with Operators *)
QAIntegrate[c_ ** p_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] :=
	c ** QAIntegrate[p, v, n, x]


QAIntegrate[p_ ** c_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] :=
	QAIntegrate[p, v, n, x] ** c


Hermitian[QAIntegrate[ex_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]] := QAIntegrate[Hermitian[ex], v, n, x]


QAIntegrate /: QAIntegrate[p_, v: (_Symbol | _DirectedInfinity), n_Integer, x__?VariableQ] ** QAIntegrate[q_, v: (_Symbol | _DirectedInfinity), n_Integer, y__?VariableQ] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QAIntegrate[p ** q, v, n, Sequence[x, y]]


(* ::Subsection:: *)
(* Summation Rules with Operators *)
QASum[c_ ** p_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] :=
	c ** QASum[p, x]


QASum[p_ ** c_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] :=
	QASum[p, x] ** c


Hermitian[QASum[ex_, x__Symbol]] := QASum[Hermitian[ex], x]


QASum /: QASum[p_, x__?VariableQ] ** QASum[q_, y__?VariableQ] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QASum[p ** q, Sequence[x, y]]


(* ::Subsection:: *)
(* Derivation Rules *)
SetAttributes[QAD, {HoldFirst, Listable}]

QAD[QAIntegrate[ex_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol], seq__] := QAIntegrate[QAD[ex, seq], v, n, x]

QAD[QASum[ex_, x__Symbol], seq__] := QASum[QAD[ex, seq], x]

QAD[fac1_ + fac2_, seq__] := QAD[fac1, seq] + QAD[fac2, seq]

QAD[fac1_ * fac2_, var_] := QAD[fac1, var] fac2 + fac1 QAD[fac2, var]

QAD[fac1_ ** fac2_, var_] := QAD[fac1, var] ** fac2 + fac1 ** QAD[fac2, var]

QAD[c_, var_] /; FreeQ[c, var] := 0

QAD[ex_, seq__] /; 
	! Head[ex] === QAIntegrate && 
	! Head[ex] === QASum && 
	! Head[ex] === Plus && 
	! Head[ex] === Times && 
	! Head[ex] === NonCommutativeMultiply := 
	D[ex, seq]


(* ::Subsection:: *)
(* Matrix Operations *)

(* Matrix operations as used in Spinor Theory *)

(* Checks for a Row Matrix *)
RowQ[m: {{__}} ? MatrixQ] := True

RowQ[m_List ? MatrixQ] := False


(* Checks for a Column Matrix *)
ColumnQ[m_List ? MatrixQ] := RowQ[Transpose[m]]


(* Makes a Row Matrix for either a Vector or a Column Matrix *)
MakeRow[m_List] /; VectorQ[m] := { m }

MakeRow[m_List] /; ColumnQ[m] := Transpose[m]

MakeRow[m_List] := m


(* Makes a Column Matrix from either a Vector or a Row Matrix *)
MakeColumn[m_List] /; VectorQ[m] := Transpose[MakeRow[m]]

MakeColumn[m_List] /; RowQ[m] := Transpose[m]

MakeColumn[m_List] := m


(* The Hermitian operation on a Matrix or a Vector
   - Matrix: apply Hermitian on each element and transpose
   - Vector: apply Hermitian on each element
*)
Hermitian[m_List] /; MatrixQ[m] := Map[Hermitian, m // Transpose, {2}]

Hermitian[m_List] /; VectorQ[m] := Map[Hermitian, m]


(* A scalar list Times an operator *)
Operator /: c_List op_Operator := Map[# * op&, c]


Unprotect[NonCommutativeMultiply];

(* The Distribution Law for NonCommutativeMultiply,
   provided summands contain QASum expressions 
*)
(smd1_ + smd2:__) ** sum_Plus /; !FreeQ[smd1, _QASum] := smd1 ** sum + (Plus@@{ smd2 }) ** sum

sum_Plus ** (smd1_ + smd2:__) /; !FreeQ[smd1, _QASum] := sum ** smd1 + sum ** (Plus@@{ smd2 })


(* The Distribution Law for NonCommutativeMultiply with List,
   provided summands contain QASum expressions 
*)
(smd1_ + smd2:__) ** mat_List /; !FreeQ[smd1, _QASum] := smd1 ** mat + (Plus@@{ smd2 }) ** mat

mat_List ** (smd1_ + smd2:__) /; !FreeQ[smd1, _QASum] := mat ** smd1 + mat ** (Plus@@{ smd2 })


(* Another form of Distribution Law for NonCommutativeMultiply between List and QASum of Lists *)
QASum[mat1_List, seq__] ** mat2_List := QASum[mat1 ** mat2, seq]

mat2_List ** QASum[mat1_List, seq__] := QASum[mat2 ** mat1, seq]


(* Matrix multiplication with matrices containing Operator(s)
   TODO: For these rules to become active Operators are required -> generalize? 
*)
mat1_List ? RowQ ** mat2_List ? VectorQ /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	mat1 ** MakeColumn[mat2]

mat1_List ? VectorQ ** mat2_List ? ColumnQ /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	MakeRow[mat1] ** mat2
	
mat1_List ? RowQ ** mat2_List ? ColumnQ /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	Inner[NonCommutativeMultiply, mat1, mat2, Plus][[1, 1]]
	
mat1_List ** mat2_List /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	Inner[NonCommutativeMultiply, mat1, mat2, Plus]
	

mat1_List ? RowQ ** mat2_List ? ColumnQ /; 
	FreeQ[mat1, _Operator] && FreeQ[mat2, _Operator] := 
	Inner[Times, mat1, mat2, Plus][[1, 1]]
	
mat1_List ** mat2_List /; 
	FreeQ[mat1, _Operator] && FreeQ[mat2, _Operator] := 
	Inner[Times, mat1, mat2, Plus]


(* Outer Product of Vectors *)
vec1_List ? VectorQ \[SmallCircle] vec2_List ? VectorQ :=
	Outer[NonCommutativeMultiply, vec1, vec2]
	 

(* NonCommutativeMultiply reduced to Times *)
fac1_ ** fac2_ /; QuantumAlgebra`Private`SimpleConstantQ[fac1] && QuantumAlgebra`Private`SimpleConstantQ[fac2] := fac1 * fac2 * Operator[1]

fac1_ ** fac2_ /; QuantumAlgebra`Private`SimpleConstantQ[fac1] || QuantumAlgebra`Private`SimpleConstantQ[fac2] := fac1 * fac2


Protect[NonCommutativeMultiply];


(* ::Subsection:: *)
(* Functions Generate Replacement Rules for (Anti-) Commutators *)

(*
	- The target ordering is either Normal Ordering or Reverse Ordering
	- The commutator is either a Anti- or Normal Commutator
	- The rule is either for Continuous or Discontinuous case 
	TODO: sign mismatches, test with Dirac Equation
*)
DefineCommutatorRule[anti_, continuous_, normal_, dim_Integer : 3] :=
	Module[{sign, head, list1, list2, op1, op2, part1, part2, comm, rule, replacement, a, b, h, k, r, s, i},
		sign = If[anti, -1, +1];
		head = If[continuous, DiracDelta, DiscreteDelta];
		list1 = Table[Subscript[k, i], {i, 1, dim}];
		list2 = Table[Subscript[h, i], {i, 1, dim}];
		op1 = Operator[a, list1, r];
		op2 = Operator[b, "H", list2, s];
		comm = head@@Inner[#1 - #2&, list2, list1, List] Operator[1];
		If[normal,
			part1 = (op1 ** op2);
			part2 = sign op2 ** op1;,
			part1 = (op2 ** op1);
			part2 = sign (op1 ** op2);
			comm = -sign comm;
		];
		replacement = Piecewise[
			{
				{ Evaluate[part2 + (* !! *) comm], Hold[SymbolName[a] == SymbolName[b] && r == s && r == {}]},
			 	{ Evaluate[part2 + comm], Hold[SymbolName[a] == SymbolName[b] && r == s]}
			}, 
			Evaluate[part2]];
		part1 = part1 /. Map[# -> Pattern[#, Blank[]]&, {a, b, h, k}];
		part1 = part1 /. Map[# -> Pattern[#, List[BlankNullSequence[]]]&, {r, s}];
		rule = 1 :> (replacement // Evaluate) // ReleaseHold;
		rule[[1]] = HoldPattern[part1 // Evaluate];
		rule
	]

(* 
	Used to bring Operators in unique order
	In the case of (Anti-) Commutating operators 
*)
DefineNameOrderingRule[anti_, dim_Integer : 3] :=
	Module[{sign, list1, list2, op1, op2, op2repl, part1, rule, replacement, a, b, h, k, r, s, n, seq, sn, i},
		sn[sy_Symbol, i2_Integer] := Symbol[SymbolName[sy] <> ToString[i2]];
		sign = If[anti, -1, +1];
		list1 = Table[Subscript[k, i], {i, 1, dim}];
		list2 = Table[Subscript[h, i], {i, 1, dim}];
		op1 = Operator[a, seq, list1, r];
		op2 = Operator[b, seq, list2, s];
		op2repl = Operator[b, seq, Table[sn[n, i], {i, 1, dim}], s];
		part1 = op1 ** op2;
		part1 = part1 /. Map[# -> Pattern[#, Blank[]]&, {a, b, k}];
		part1 = part1 /. Map[Subscript[h, #] -> Pattern[sn[n, #] // Evaluate, Subscript[h_, #] | -Subscript[h_, #]]&, Range[dim]];
		part1 = part1 /. seq -> seq___;
		part1 = part1 /. Map[# -> Pattern[#, List[BlankNullSequence[]]]&, {r, s}];
		replacement = 
			If[Hold[Order[SymbolName[a], SymbolName[b]] >= 0],
				Evaluate[op1 ** op2repl],
				Evaluate[sign op2repl ** op1]
			];
		rule = 1 :> (replacement // Evaluate) // ReleaseHold;
		rule[[1]] = HoldPattern[part1 // Evaluate];
		rule
	]


(* ::Subsection:: *)
(* Sort negative indices in operator product *)
QASum[Times[facs1__, Operator[a_Symbol, seq___, lst1: {__}, r: {___}] ** Operator[b_Symbol, seq___, lst2: {__}, s: {___}]], k_Symbol] /;
	Module[{i},
		MatchQ[lst1, Table[-Subscript[k, i], {i, 1, Length[lst1]}]] &&
		MatchQ[lst2, Table[Subscript[k, i], {i, 1, Length[lst1]}]] &&
		({ seq } == {} || seq == "H")
	]:=
	Module[{},
		QASum[Times[
			Sequence@@({ facs1 } /. Subscript[k, i_] /; 1 <= i <= Length[lst1] -> -Subscript[k, i]),
			Operator[a, seq, lst2, r] ** Operator[b, seq, lst1, s]],
			k]
	]

(* ::Subsection:: *)
(* Reverse Application of the Distribution Law for Operators or Operator Products with scalar factors *)
Unprotect[Plus]

Plus /: 
	c_ Operator[a_, seqa___, lsta: {___}, r: {___}] ** Operator[b_, seqb___, lstb: {___}, s: {___}] + 
	d_ Operator[a_, seqa___, lsta: {___}, r: {___}] ** Operator[b_, seqb___, lstb: {___}, s: {___}] := 
	(c + d) Operator[a, seqa, lsta, r] ** Operator[b, seqb, lstb, s]

Plus /: c_ Operator[a_] ** Operator[b_] + d_ Operator[a_] ** Operator[b_] := (c + d) Operator[a] ** Operator[b]

Plus /: 
	c_ Operator[a_, seqa___, lsta: {___}, r: {___}] + d_ Operator[a_, seqa___, lsta: {___}, r: {___}] := 
	(c + d) Operator[a, seqa, lsta, r]

Plus /: c_ Operator[a_] + d_ Operator[a_] := (c + d) Operator[a]

Protect[Plus]


(* ::Subsection:: *)
(* Variables *)
replacementRules = <| "default" -> {}, "empty" -> {} |>
variables = <| "default" -> {}, "empty" -> {} |>
timeConstraint = <| "default" -> 300. |>
assumptions = <| "default" -> {}, "empty" -> {} |>
default = "default"

getSet[set_String] := set

getSet[Null] := default


(* ::Subsection:: *)
(* Replacement Rules *)
QAReplace[ex_, rules_String] := ex /. replacementRules[rules]

QAReplace[ex_, rules: { __String }] := ex /. Dispatch[Flatten[Map[replacementRules[#]&, rules]]]

QAReplace[rules: _String | { __String }][ex_] := QAReplace[ex, rules] 

QAReplaceRepeated[ex_, rules_String] := ex //. replacementRules[rules]

QAReplaceRepeated[ex_, rules: { __String }] := ex //. Dispatch[Flatten[Map[replacementRules[#]&, rules]]]

QAReplaceRepeated[rules: _String | { __String }][ex_] := QAReplaceRepeated[ex, rules] 


(* TODO: remove unnecessary replacement rules, IMPROVE SORT VARIABLES
   nameOrdering, NormalOrdering, ReverseOrdering, NormalOrderingContinuous, ReverseOrderingContinuous
   IntegrateContract.old, SumContract.old => all removed after check-in
*)
QAInitializeReplacementRules[] :=
	Module[{},
		(* Integrate about Sums (Plus operation) is splitted *)
		QAAddReplacementRules[
			{
				HoldPattern[QAIntegrate[p_ + q_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]] :> (QAIntegrate[p, v, n, x] + QAIntegrate[q, v, n, x]),
				HoldPattern[QAIntegrate[p_ - q_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]] :> (QAIntegrate[p, v, n, x] - QAIntegrate[q, v, n, x])			
			},
			"IntegrateExpand"];

		(* Sum of Integrates is joined *)
		QAAddReplacementRules[
			{
				HoldPattern[((Times[c___, QAIntegrate[p_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]] | QAIntegrate[p_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]) + (Times[d___, QAIntegrate[q_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]] | QAIntegrate[q_, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]))] :> 
					QAIntegrate[Times[c, p] + Times[d, q], v, n, x]
			},
			"IntegrateContract"];
		
		(* Sum about Sums (Plus operation) is splitted *)
		QAAddReplacementRules[
			{
				HoldPattern[QASum[p_ + q_, x__Symbol]] :> (QASum[p, x] + QASum[q, x]),
				HoldPattern[QASum[p_ - q_, x__Symbol]] :> (QASum[p, x] - QASum[q, x])			
			},
			"SumExpand"];
		
		(* Sum of Sums is joined *)
		QAAddReplacementRules[
			{
				HoldPattern[((Times[c___, QASum[p_, x__Symbol]] | QASum[p_, x__Symbol]) + (Times[d___, QASum[q_, x__Symbol]] | QASum[q_, x__Symbol]))] :> 
					QASum[Times[c, p] + Times[d, q], x]
			},
			"SumContract"];
		
		(* The normal distribution law *)
		QAAddReplacementRules[
			{
				HoldPattern[Times[fac__, sum_Plus]] :> Map[Times[fac, #]&, sum]
			},
			"DistributionLaw"];
		
		(* Order of variables in certain constructs like Integrate, Sum, DiracDelta, DiscreteDelta is sorted by Variable Name *)
		QAAddReplacementRules[
			{
				HoldPattern[QAIntegrate[ex_, v: (_Symbol | _DirectedInfinity), n_Integer, seq__?VariableQ]] :> 
					QAIntegrate[ex, v, n, Sequence@@Map[Symbol, Sort[Map[SymbolName, {seq}]]]],
				
				HoldPattern[QASum[ex_, seq__?VariableQ]] :> 
					QASum[ex, Sequence@@Map[Symbol, Sort[Map[SymbolName, {seq}]]]],
				
				HoldPattern[DiracDelta[Subscript[h_, 1] - Subscript[k_, 1], Subscript[h_, 2] - Subscript[k_, 2], Subscript[h_, 3] - Subscript[k_, 3]]] :> 
					If[
						Order[SymbolName[h], SymbolName[k]] >= 0, 
						DiracDelta[Subscript[h, 1] - Subscript[k, 1], Subscript[h, 2] - Subscript[k, 2], Subscript[h, 3] - Subscript[k, 3]], 
						DiracDelta[Subscript[k, 1] - Subscript[h, 1], Subscript[k, 2] - Subscript[h, 2], Subscript[k, 3] - Subscript[h, 3]]
					],
				
				HoldPattern[DiscreteDelta[Subscript[h_, 1] - Subscript[k_, 1], Subscript[h_, 2] - Subscript[k_, 2], Subscript[h_, 3] - Subscript[k_, 3]]] :> 
					If[
						Order[SymbolName[h], SymbolName[k]] >= 0, 
						DiscreteDelta[Subscript[h, 1] - Subscript[k, 1], Subscript[h, 2] - Subscript[k, 2], Subscript[h, 3] - Subscript[k, 3]], 
						DiscreteDelta[Subscript[k, 1] - Subscript[h, 1], Subscript[k, 2] - Subscript[h, 2], Subscript[k, 3] - Subscript[h, 3]]
					]			
			},
			"SortVariables"];
	True]


QAResetReplacementRules[set_String : Null] := AssociateTo[replacementRules, getSet[set] -> {}]


QAAddReplacementRules[rules_List, set_String : Null] :=
	Module[{old, new},
		old = Lookup[replacementRules, getSet[set], {}];
		new = Join[old, rules];
		AssociateTo[replacementRules, getSet[set] -> new]
	]


(* ::Subsection:: *)
(* Integration Variables *)

QAResetExplicitIntegrationVariables[set_String : Null] := AssociateTo[variables, getSet[set] -> {}]


QAAddExplicitIntegrationVariables[vars_List, set_String : Null] :=
	Module[{old, new},
		old = Lookup[variables, getSet[set], {}];
		new = Join[old, vars];
		AssociateTo[variables, getSet[set] -> new]
	]


(* ::Subsection:: *)
(* Assumption Handling *)

QASelectDefault[set_String : "default"] := (default = set)


QAResetAssumptions[set_String : Null] := AssociateTo[assumptions, getSet[set] -> {}]


QAAddAssumptions[ass_List, set_String : Null] :=
	Module[{old, new},
		old = Lookup[assumptions, getSet[set], {}];
		new = Join[old, ass];
		AssociateTo[assumptions, getSet[set] -> new]
	]


QASetTimeConstraint[constraint_, set_String : Null] := AssociateTo[timeConstraint, getSet[set] -> constraint]

QASimplify[ex_, set_String] := 
	Simplify[ex, 
		Assumptions -> Map[And, assumptions[getSet[set]]], 
		TimeConstraint -> Lookup[timeConstraint, getSet[set], 300.]
	]

QASimplify[set_String : Null][ex_] := QASimplify[ex, getSet[set]]

End[]

EndPackage[]
